; T64Loader
; A replacement for Novaload and others to load from
; a specially prepared Magic Desk cartridge.
; Written in Feburary 2024 by Boris Schneider-Johne / docbobo
; Feel free to reuse and improve, send me a link
; to the stuff you build with this: boris@dreisechzig.net

!TO "t64loader.bin",plain

RAMSTART = $033C        ; the RAM part of the loader goes here
COMPILEFORBANK = $00    ; change here for the loader to live in a different bank
ENABLEAUTOSTART = 1     ; compile for autostart
ENABLEDIRECTORY = 1     ; compile directory function
COLORRAMFIX = 1         ; reroute all loads that are only for $d800 to $dbff to colorram
FLICKER = 1             ; if you want slow border flicker (only every 2K so visible only in long loads)

* = $8000                ; ROM starts at $8000

; Zeropage Vectors and Adresses, all Kernal LOAD compatible
ZPLOADVERIFY1 = $0A ;00 for Load, all others Verify

ZPLOADVERIFY2 = $93 ;00 for Load, all others Verify, USABLE because we never verify and it is set by ACC at kernal load entry
CURRENTBANK = $93   ;reusing the Verify byte to store the current bank before and during copy

ZPENDPROGRAM = $2D  ;End of loaded program plus 1

ZPDEVICENR = $49    ;Device Number for LOAD and SAVE

ZPNAMELENGTH = $b7  ;length of filename

ZPFILENAME = $bb    ;pointer to filename
ZPDIRENTRY = $bb    ;pointer to our directory entry after we finished the search

ZPSECONDADDR = $b9  ;Secondary Adress, use $c3 if not Zero
ZPLOADADRESS2 = $c3 ;adress to load to when secondary is 0 ($0801 or set by code)

ZPDEVICENR2 = $ba   ;Device Number for current file
BUFFERBYTE = $ba    ;We ignore the device number and catch them all, this allows use as $01 buffer

ZPMOTORSWITCH =$c0  ;00 Datasette Motor is off, others on

ZPLOADADRESS = $ae  ;Load Adress; rolling with the load
ZPDIRSEARCH = $ae   ;BEFORE actual load re-use for directory search


; Vectors And Adresses in the $300 space
OPENFILE = $031A
CLOSEFILE = $031C
LOADVECTOR = $0330
SAVEVECTOR = $0332

; Kernal routines
PRINTSTRING = $AB1E
KERNALFILENOTFOUND =$F704

; Needed offsets in the T64 Directory

DIRECTORYSTART = TAPESTART+$40  ; t64 starts at $8x00, first 32 bytes are header
NUMBEROFFILES = TAPESTART+$22   ; Number of directory entries
OFFSETLOADADDR = $02    ; Where to load the file unless secondary adress!
OFFSETENDADDR = $04     ; EndAdress
OFFSETFILENAME = $10    ; Filename starts here padded with Space
OFFSETT64 = $08         ; Start of ROM in file (PRG 2-byte Header has been chopped off!)


;--------------------------------------------------------

; Boot Vectors
!word COLDSTART
!word WARMSTART ;points to Loader Install - can be used as entry point for other software

; Cartridge Signature
!byte $c3,$c2,$cd,$38,$30

!IF ENABLEDIRECTORY {
TEXTSTRING

!text "T64CRT"    
!byte $00
}

;--------------------------------------------------------

COLDSTART

; Kernal Reset Routine
  SEI
  STX $D016   ; Turn on VIC for PAL / NTSC check
  JSR $FDA3   ; IOINIT - Init CIA chips
  JSR $FD50   ; RANTAM - Clear/test system RAM
  LDA #$A0
  STA $0284   ; ignore cartridge ROM for end of detected RAM for BASIC
  JSR $FD15   ; RESTOR - Init KERNAL RAM vectors
  JSR $FF5B   ; CINT   - Init VIC and screen editor
  
  LDA #$69    ; make sure that BRK does not clear all vectors by jumping one ahead
  STA $0316
  

; We need to patch this code back to normal because Autoload has changed this adress, making subsequent loads fail!

  JSR COPYLOADER
  

  

!IF ENABLEAUTOSTART {

; Autostart the first file when Filename starts with !
  LDA DIRECTORYSTART+OFFSETFILENAME   
  CMP #$21
  BNE BASIC                           ;go to BASIC 
  LDA #$00
  STA ZPNAMELENGTH                    ;if name length is zero, loader will load the first file (shortcut)
  
  LDA DIRECTORYSTART+OFFSETFILENAME+4
  JSR PETSCIITOBYTE  
  STA JUMPAWAY+1
  LDA DIRECTORYSTART+OFFSETFILENAME+3
  JSR PETSCIITOBYTE
  ASL
  ASL
  ASL
  ASL
  ADC JUMPAWAY+1
  STA JUMPAWAY+1
  
  LDA DIRECTORYSTART+OFFSETFILENAME+2
  JSR PETSCIITOBYTE
  STA JUMPAWAY+2
  LDA DIRECTORYSTART+OFFSETFILENAME+1
  JSR PETSCIITOBYTE
  ASL
  ASL
  ASL
  ASL
  ADC JUMPAWAY+2
  STA JUMPAWAY+2

  JSR $F4A2                     ;Load the file with our routine JMP ($033C)
                                ;has been patched to JMP into the filename adress

PETSCIITOBYTE
  SEC
  SBC #$30
  CMP #$0A
  BCC +
  SBC #$07
  +
  RTS

}

BASIC

!IF ENABLEDIRECTORY {
; BASIC RESET  Routine
  JSR $E453   ; Init BASIC RAM vectors
  JSR $E3BF   ; Main BASIC RAM Init routine
  JSR $E422   ; Power-up message / NEW command

  LDA #<TEXTSTRING
  LDY #>TEXTSTRING
  JSR PRINTSTRING ;add short message to show loader is activated

  LDA #$37
  STA BUFFERBYTE  ;make ROM ON the default
 
  JMP JUMPTOBASIC
;------------------------------------ End of Coldstart Routine
}

WARMSTART

COPYLOADER
  LDY #ENDOFLOADER-RAMSTART+1
-
  LDA LOADERROMADDR-1,Y
  STA RAMSTART-1,Y
  DEY
  BNE -

; Set Loadvector to our routine and Savevector to CLC and Return (do nothing)
  LDA #<LOADENTRY
  STA LOADVECTOR
  LDA #>LOADENTRY
  STA LOADVECTOR+1
  
  LDA #<COPYENDED
  STA SAVEVECTOR
  LDA #>COPYENDED
  STA SAVEVECTOR+1
  RTS

;---------------------------------- End of warmstart routine

T64CODE ; We are coming here from the Low memory loader (see below). To keep loader short, code that does not need to be modified stays in ROM.
;------

  LDA ZPNAMELENGTH  ;when name has length zero, replace with first file
  BNE +
  INC ZPNAMELENGTH
  JMP REPLACEFIRSTLETTER
+

  LDA #<STAYHERE  ; need to fix the JMP back to do nothing if we changed it for Autoload!
  STA JUMPAWAY+1  ; Autoload works with Filename Zero, so if there is a file name we patch back
  LDA #>STAYHERE  ; Does work in edge case too because if we manually load, we never patched
  STA JUMPAWAY+2

  
  LDA FILENAMECOPY
!IF ENABLEDIRECTORY {  
  CMP #"$"       ;first character $ then go to dirprint
  BEQ DIRPRINT
}
  
  CMP #"*"      ;replace * in Pos 0 with first filename in Pos 0 to fool Directory Search
  BNE NOTFIRST
  
REPLACEFIRSTLETTER
  LDA DIRECTORYSTART+OFFSETFILENAME
  STA FILENAMECOPY

NOTFIRST
  JMP NOTDIR

!IF ENABLEDIRECTORY {
 
DIRPRINT    ; Print first 7 directory Entries direct to screen
  LDY #$00
LOOPDIR
  LDX #$00
-
  LDA DIRECTORYSTART+OFFSETFILENAME,Y
  STA DIRECTORYSTRING,X
  INX
  INY
  TYA
  CMP #$E0
  BCS DIRDONE
  PHA
  TXA 
  CMP #$0F
  BNE -
  LDA #<DIRECTORYSTRING
  LDY #>DIRECTORYSTRING
  JSR PRINTSTRING
  PLA
  CLC
  ADC #$11
  TAY
  BNE LOOPDIR
DIRDONE
  JMP JUMPTOBASIC
  
}  

NOTDIR
  LDX NUMBEROFFILES ; counting down with X for "file not found"
  LDA #>DIRECTORYSTART
  STA ZPDIRSEARCH+1
  LDA #<DIRECTORYSTART
  CLC
  ADC #$10
  STA ZPDIRSEARCH
  BCC +
  INC ZPDIRSEARCH+1
 +
  
NAMECHECK ; check filenames starting at last character as T64 pads with spaces

  LDY ZPNAMELENGTH
  DEY
-
  LDA (ZPDIRSEARCH),Y
  CMP FILENAMECOPY,Y
  BNE NEXTFILE
  DEY
  BPL -
  JMP FOUNDFILE
  
NEXTFILE
  DEX
  BEQ FILENOTFOUND
  LDA ZPDIRSEARCH
  CLC
  ADC #$20
  STA ZPDIRSEARCH
  BCC +
  INC ZPDIRSEARCH+1
+
  JMP NAMECHECK

FILENOTFOUND ; need to jump out to file not found!
  JMP FILENOTFOUNDEXIT
  
FOUNDFILE
; We now reuse the pointer to the file name as the pointer to the directory entry and can store the load adress
  LDA ZPDIRSEARCH
  SEC 
  SBC #$10
  STA ZPDIRENTRY
  LDA ZPDIRSEARCH+1
  SBC #$00
  STA ZPDIRENTRY+1

;Get LOADADRESS in RAM (TO:) and store directly in code
  LDY #OFFSETLOADADDR
  LDA (ZPDIRENTRY),Y
  STA TORAM+1
  INY
  LDA (ZPDIRENTRY),Y
  STA TORAM+2
  CMP #$05            ; do not Load below $0400
  BCS +
  JMP ERROR
  +

; Store End Adresss in the code for the byte copy
  LDY #OFFSETENDADDR
  LDA (ZPDIRENTRY),Y
  STA ENDCOUNTLO+1
  INY
  LDA (ZPDIRENTRY),Y
  STA ENDCOUNTHI+1 

; if secondary adress is set, we need to change start and end adress
; we calculate the new end adress first

  LDA ZPSECONDADDR
  BEQ GETSTART   
  
  SEC
  LDA ENDCOUNTLO+1
  SBC TORAM+1
  TAX
  LDA ENDCOUNTHI+1
  SBC TORAM+2
  TAY
  TXA
  CLC
  ADC ZPLOADADRESS2
  STA ENDCOUNTLO+1
  TYA
  ADC ZPLOADADRESS2+1
  STA ENDCOUNTHI+1
  
; now we replace the load adress with the one given in by kernal/call
  LDA ZPLOADADRESS2
  STA TORAM+1
  LDA ZPLOADADRESS2+1
  STA TORAM+2  
  
GETSTART

; Get STARTADRESS in ROM (FROM:) by converting 3-Byte position into bank and position
; The startadress inside the bank is written directly into the copy routine which is self-modifying

  LDY #OFFSETT64  
  LDA (ZPDIRENTRY),Y
  STA FROMCRT+1     ;low byte of adress stays untouched
  INY
  INY
  LDA (ZPDIRENTRY),Y  ;High byte as start value for correct bank calculation
  CLC
  ROL
  ROL                 ;Multiply with 8 (8 Banks per 64K)
  ROL
  ADC #COMPILEFORBANK ;Add the bank the loader is compiled for 
  STA CURRENTBANK     ;and store it.
  DEY
  LDA (ZPDIRENTRY),Y  ;Get the mid byte
  CLC
  ADC #>TAPESTART-$80 ;add $0x00 as directory starts at $8x00 in the first bank
  BCC +
  INC CURRENTBANK     ;if we rolled over (carry set), we need to increment the bank
+

  LSR                 ;divide the mid byte by 32 (as we have 8 banks)
  LSR
  LSR
  LSR
  LSR
  CLC
  ADC CURRENTBANK     ;add this to the current high byte to get to the correct bank
  STA CURRENTBANK
  LDA (ZPDIRENTRY),Y  ;lets get the mid byte again and put into $8000 to $9FFF range
  CLC
  ADC #>TAPESTART-$80 ; add $0x00 as directory starts at $8x00 in the first bank - bank switch happened above
  AND #$9F
  ORA #$80
  STA FROMCRT+2       ;conversion is done

!if COLORRAMFIX = 1 {  ; if the load start and end adress are only in color ram space, patch the copy routine to NOT write to RAM. It repairs itself at the end.
  LDA TORAM+2          ; can be turned off during compile to save a few bytes if you KNOW the game does not load straight into color ram.
  CMP #$D8
  BCC NOCOLORRAM
  CMP #$DC
  BCS NOCOLORRAM
  LDA ENDCOUNTHI+1
  CMP #$DC
  BCS NOCOLORRAM 
  LDA #$37
  STA CRFIX1+1
  STA CRFIX2+1
NOCOLORRAM
}
  
;All done in ROM now switch into RAM for the actual copy

  JMP MOVESTART

; File not found and load below $0400 go here
ERROR
  SEC
  JMP GOODBYEERROR

;----------THE END-------------------------

; Start of the CRT2RAM routine, usually at $033c but fully movable to RAMSTART!

LOADERROMADDR
!PSEUDOPC RAMSTART
LOADENTRY

  SEI
; Save Memory Status
  LDA $01
  STA BUFFERBYTE

; Need to copy filename into bufferbytes at the end (doubling up for directory print)
; because the filename might be under the ROM

  LDY #$00
-
  CPY ZPNAMELENGTH
  BEQ +  
  LDA (ZPFILENAME),Y
  STA FILENAMECOPY,Y
  INY
  BNE -
+  

  LDA #$37
  STA $01
  LDA #COMPILEFORBANK ;switch to the bank the loader is in
  STA $DE00
  JMP T64CODE ; JMP to all code that does not need to live down here, then we return to the next address


MOVESTART
;--------

; Switch to the first bank our file is in 

  LDX CURRENTBANK
  STX $DE00
CRFIX1
  LDA #$33
  STA $01

COPYLOOP 
  FROMCRT     ;selfmodify in several places
  LDA $8000   ;selfmodify
  TORAM       ;selfmodify
  STA $0801   ;selfmodify
  
  INC FROMCRT+1 ;increment from adress, when reaching $A000, bankswitch and back to $8000
  BNE +
  INC FROMCRT+2
  LDA FROMCRT+2
  CMP #$A0
  BNE +
  LDA #$37
  STA $01
  INX
  STX $DE00

!IF FLICKER=1 {
  INC $D020
}   
  
CRFIX2
  LDA #$33
  STA $01
  LDA #$80
  STA FROMCRT+2
+

  INC TORAM+1 ;increment to adress
  BNE +
  INC TORAM+2
+

  LDA TORAM+1   ;compare current adress with T64 end adress (which is +1 so we do not want to copy this byte!), if equal we are done
ENDCOUNTLO      ;selfmodified code, lo byte end adress goes here
  CMP #$00
  BNE COPYLOOP
  LDA TORAM+2
ENDCOUNTHI      ;selfmodified code, hi byte end adress goes here
  CMP #$00
  BNE COPYLOOP 

COPYENDED 

!if COLORRAMFIX = 1 { ; This repairs our code always because we MIGHT have patched here to load into colorram
  LDA #$33
  STA CRFIX1+1
  STA CRFIX2+1
}

  LDA #<STAYHERE  ; need to fix this if we changed for Autoload!
  STA STAYHERE+1
  LDA #>STAYHERE
  STA STAYHERE+2
  
  CLC       ; important, carry flag means error

GOODBYEERROR
CARTOFF
  LDA #$37 
  STA $01
  LDA #$80      ;switch off Magic Cart
  STA $DE00    
  LDA BUFFERBYTE
  STA $01
  CLI
JUMPAWAY
  JMP STAYHERE ; will be modified to the real JMP adress for autoload
STAYHERE
  RTS

JUMPTOBASIC
  JSR CARTOFF
  JMP $A7AE
  
FILENOTFOUNDEXIT
  JSR CARTOFF
  JMP KERNALFILENOTFOUND
  
FILENAMECOPY
!IF ENABLEDIRECTORY {
DIRECTORYSTRING
}
!text "                "
;      1234567890123456
!byte $0D,$00  
 
ENDOFLOADER

!REALPC
!realign 256
; T64 starts here!
TAPESTART